import urwid
from .ZeitzonoTime import ZeitzonoTime
from .ZeitzonoCities import ZeitzonoCities
from .ZeitzonoUrwidSearch import ZeitzonoUrwidSearch
from .ZeitzonoSearch import ZeitzonoSearch
from .ZeitzonoDB import ZeitzonoDB


class ZeitzonoUrwidMain(urwid.WidgetWrap):
    _selectable = True

    def __init__(self, loop, zeitzonowidgetswitcher, cache, version):

        self.zeitzonotime = ZeitzonoTime()
        self.zeitzonocities = ZeitzonoCities()
        self.zeitzonodb = ZeitzonoDB()
        self.footerpile = None
        self.nlp_mode = False  # if set, we are parsing keyboard input differently
        self.clock_mode = True  # if set, we update the base clock in realtime
        self.loop = loop  # this is needed call set_alarm_in()
        self.index_mode = False  # if set, we are picking an item in list

        # if we are caching, read HoronoCities from cache
        self.cache = cache
        if cache is not None:
            try:
                with open(self.cache) as cachefile:
                    self.zeitzonocities.fromJSON(cachefile)
            except Exception:
                pass

        self.baset = urwid.Text("", wrap="clip", align="left")
        basetmap = urwid.AttrMap(self.baset, "main_base_time")
        self.basezc = urwid.Text("", wrap="clip", align="right")
        basezcmap = urwid.AttrMap(self.basezc, "main_basezc")

        self.basezc_is_c = False
        self.basezc_city = None
        self.baset_update()
        self.basezc_update()

        htext = "zeitzono "
        htext_len = len(htext)
        zeitzono_ut = urwid.Text(htext, wrap="clip", align="right")
        zeitzono_ut_am = urwid.AttrMap(zeitzono_ut, "main_zeitzono")

        self.version = version
        version_len = len(version)
        version_ut = urwid.Text(version, wrap="clip", align="right")
        version_ut_am = urwid.AttrMap(version_ut, "main_version")

        blank = urwid.Text("", align="right")
        versioncols = urwid.Columns(
            [
                ("weight", 99, blank),
                (htext_len, zeitzono_ut_am),
                (version_len, version_ut_am),
            ]
        )

        self.bodypile = urwid.Pile(self.body_gen())

        self.bodyfill = urwid.Filler(self.bodypile, valign="bottom")
        self.zeitzonowidgetswitcher = zeitzonowidgetswitcher

        cols = urwid.Columns([basetmap, basezcmap])
        blankline = urwid.Text("", wrap="clip")

        helpline = "? - help,  c - add cities, n - display current time, Q - quit"
        helpline = urwid.Text(helpline, wrap="clip")
        helpline_attr = urwid.AttrMap(helpline, "main_helpline")

        self.footer = [blankline, helpline_attr, cols]
        self.footerpile = urwid.Pile(self.footer, focus_item=len(self.footer) - 1)
        self.footer_bak = [(x, ("pack", None)) for x in self.footer]

        # ---------------------------------
        # create a footer pile for nlp mode

        helpline_nlp = "enter a human-readable date/time string:"
        helpline_nlp = urwid.Text(helpline_nlp, wrap="clip", align="left")
        helpline_nlp_attr = urwid.AttrMap(helpline_nlp, "main_helpline")
        self.nlp_prompt = urwid.Edit(caption="(NLP) zeitzono> ", align="left")
        nlp_prompt_attr = urwid.AttrMap(self.nlp_prompt, "search_prompt")
        nlp_pilelist = [blankline, helpline_nlp_attr, nlp_prompt_attr]
        self.nlp_pilelist = [(x, ("pack", None)) for x in nlp_pilelist]

        # ---------------------------------

        frame = urwid.Frame(self.bodyfill, header=versioncols,
                            footer=self.footerpile, focus_part='footer')

        super().__init__(frame)

    def _list_is_max_capacity(self, fakecap=None):
        cols, rows = urwid.raw_display.Screen().get_cols_rows()
        maxrows = rows - 4
        cap = self.zeitzonocities.numcities()

        if fakecap is not None:
            cap = cap + fakecap - 1

        if cap >= maxrows:
            return True

        return False

    def time_adjust(self, key):  # noqa

        if key in "SsMmHhXxFfDdWwOoYy0":
            self.clock_mode = False
            if key in ("S"):
                self.zeitzonotime.sub_sec()
            if key in ("s"):
                self.zeitzonotime.add_sec()
            if key in ("M"):
                self.zeitzonotime.sub_min()
            if key in ("m"):
                self.zeitzonotime.add_min()
            if key in ("H"):
                self.zeitzonotime.sub_hour()
            if key in ("h"):
                self.zeitzonotime.add_hour()
            if key in ("X"):
                self.zeitzonotime.sub_qday()
            if key in ("x"):
                self.zeitzonotime.add_qday()
            if key in ("F"):
                self.zeitzonotime.sub_qhour()
            if key in ("f"):
                self.zeitzonotime.add_qhour()
            if key in ("D"):
                self.zeitzonotime.sub_day()
            if key in ("d"):
                self.zeitzonotime.add_day()
            if key in ("W"):
                self.zeitzonotime.sub_week()
            if key in ("w"):
                self.zeitzonotime.add_week()
            if key in ("O"):
                self.zeitzonotime.sub_month()
            if key in ("o"):
                self.zeitzonotime.add_month()
            if key in ("Y"):
                self.zeitzonotime.sub_year()
            if key in ("y"):
                self.zeitzonotime.add_year()
            if key in ("0"):
                self.zeitzonotime.zero_sec()
                self.zeitzonotime.zero_min()
        if key in ("n"):
            self.zeitzonotime.set_time_now()
            self.clock_mode = True
            self.clock_update(self.loop, None)
        self.body_render()
        self.baset_update()
        return True

    def keypress(self, size, key):  # noqa

        if self.index_mode:
            index = self.label2index(key)
            if index is None:
                self.index_mode_off()
                return True
            else:

                if self.index_mode_type == "zone":
                    city = self.zeitzonocities.cities[index]
                    self.basezc_is_c = True
                    self.basezc_city = city.name
                    self.zeitzonotime.set_tz(city.tz)
                    self.basezc_update()

                if self.index_mode_type == "pop":
                    self.zeitzonocities.del_index(index)

                self.index_mode_off()
                return True

        if self.nlp_mode:
            self.nlp_prompt.keypress((size[0],), key)
            if key == "enter":
                x = self.nlp_prompt.get_edit_text()

                # exit if nothing entered
                if not x.strip():
                    self.nlp_prompt.set_edit_text("")
                    self.nlp_mode = False
                    self.footerpile.contents[:] = self.footer_bak

                # try and set time
                success = self.zeitzonotime.set_time_nlp(x)

                # if it doesn't work, try again
                if not success:
                    self.nlp_prompt.set_edit_text("")
                    return

                # if it does work, set clock and exit
                self.nlp_prompt.set_edit_text("")
                self.nlp_mode = False
                self.footerpile.contents[:] = self.footer_bak
                self.body_render()
                self.baset_update()
                return True
            return
        if key.lower() in ("q"):
            # if we are caching, write HoronoCities to cache before exiting
            if self.cache is not None:
                with open(self.cache, "w") as cachefile:
                    self.zeitzonocities.toJSON(cachefile)
            raise urwid.ExitMainLoop()
        if key in ("N"):
            self.clock_mode = False
            self.nlp_mode = True
            self.footerpile.contents[:] = self.nlp_pilelist
        if key == ("?"):
            self.zeitzonowidgetswitcher.switch_widget_help_main()
        if key in ("C"):
            self.zeitzonocities.clear()
        if key in ("c") and not self._list_is_max_capacity():
            zeitzonourwidsearch = ZeitzonoUrwidSearch(
                self.zeitzonowidgetswitcher,
                self.zeitzonocities,
                self.version,
                self.zeitzonodb,
            )
            self.zeitzonowidgetswitcher.set_widget("search", zeitzonourwidsearch)
            self.zeitzonowidgetswitcher.switch_widget_search()
        if key in ("L"):
            self.zeitzonotime.set_tz_local()
            self.basezc_is_c = False
            self.basezc_update()
            return True
        if key in ("p"):
            self.zeitzonocities.del_first()
            self.body_render()
            return True
        if key in ("P"):
            if self.zeitzonocities.numcities() > 0:
                self.index_mode_on(mode="pop")
            return True
        if key in ("."):
            self.zeitzonocities.rotate_right()
            self.body_render()
            return True
        if key in (","):
            self.zeitzonocities.rotate_left()
            self.body_render()
            return True
        if key in ("/"):
            self.zeitzonocities.roll_2()
            self.body_render()
            return True
        if key in ("'"):
            self.zeitzonocities.roll_3()
            self.body_render()
            return True
        if key in (";"):
            self.zeitzonocities.roll_4()
            self.body_render()
            return True
        if key in ("]"):
            self.zeitzonocities.sort_utc_offset()
            self.body_render()
            return True
        if key in ("["):
            self.zeitzonocities.sort_utc_offset(reverse=True)
            self.body_render()
            return True
        if key in ("z"):
            if self.zeitzonocities.numcities() > 0:
                city = self.zeitzonocities.cities[0]
                self.basezc_is_c = True
                self.basezc_city = city.name
                self.zeitzonotime.set_tz(city.tz)
                self.basezc_update()
            return True
        if key in ("Z"):
            if self.zeitzonocities.numcities() > 0:
                self.index_mode_on(mode="zone")
            return True
        if key in ("u"):
            self.zeitzonocities.undo()
        if key in ("r"):
            self.zeitzonocities.redo()
        if key in ("-"):
            if self._list_is_max_capacity():
                return True
            zsearch = ZeitzonoSearch(db=self.zeitzonodb)
            rcities = zsearch.random1()
            self.zeitzonocities.addcities(rcities)
            self.body_render()
            return True
        if key in ("="):
            if self._list_is_max_capacity(fakecap=10):
                return True
            zsearch = ZeitzonoSearch(db=self.zeitzonodb)
            rcities = zsearch.random10()
            self.zeitzonocities.addcities(rcities)
            self.body_render()
            return True
        self.time_adjust(key)
        self.body_render()

    def baset_update(self):

        # If we have a terminal that is sized ~80 columns wide (eg. 80x25),
        # adding the [C] in wall clock mode will truncate part of the base
        # time.
        #
        # So to get around that awkwardness, if we are smaller than some
        # arbitrary number of columns, we don't print the "base time: " string.
        cols, _ = urwid.raw_display.Screen().get_cols_rows()
        newstr = "base time: "
        if cols < 95:
            newstr = ""

        newstr = newstr + "%s" % self.zeitzonotime.get_time_str()
        newstr = [("main_base_time", newstr)]

        if self.clock_mode:
            newstr = newstr + [("main_base_time_current", " [C]")]
        self.baset.set_text(newstr)

    def basezc_update(self):
        newstr = "base zone: %s" % str(self.zeitzonotime.get_tz())
        if self.basezc_is_c:
            newstr = "base city: %s" % self.basezc_city
        self.basezc.set_text(newstr)
        self.baset_update()

    def body_gen(self, update=False, nourwid=False):

        bodylist = []
        for index, city in enumerate(self.zeitzonocities.cities):
            bodytext = str(self.zeitzonotime.get_time_tz_str(city.get_tz()))
            bodytext = bodytext + " "
            bodytext = bodytext + str(city)

            # if we are in indexmode, prepend a 0-9a-zA-Z to
            # the bottom 62 items on stack
            if self.index_mode:
                if index < 62:
                    index_label = self.index2label(index) + ": "
                else:
                    index_label = "   "
                # index_label_u = urwid.Text(index_label, wrap="clip", align="left")
                # index_label_a = urwid.AttrMap(index_label_u, "main_index")

            # nourwid is true when in --list-cached mode
            if nourwid:
                rawtext = bodytext
            else:
                if self.index_mode:
                    rawtext = urwid.Text([("main_index", index_label), ("default", bodytext)], wrap="clip", align="left")
                else:

                    rawtext = urwid.Text(bodytext, wrap="clip", align="left")
            if nourwid:
                content = rawtext
            else:
                if update:
                    content = (rawtext, ("pack", None))
                else:
                    content = rawtext

            bodylist.append(content)
        bodylist.reverse()

        return bodylist

    def body_render(self):
        self.bodypile.contents = self.body_gen(update=True)

    def clock_update(self, loop, user_date=None):

        # we only update in clock mode
        if not self.clock_mode:
            return

        # update to current time
        self.zeitzonotime.set_time_now()

        # update clock
        self.baset_update()

        # update body
        self.body_render()

        # how long to wait between refreshes, in seconds
        sleeptime = 1

        # set alarm for next refresh
        self.loop.set_alarm_in(sleeptime, self.clock_update)

    def index_mode_on(self, mode=None):
        self.index_mode_type = mode
        self.index_mode = True
        self.body_render()

    def index_mode_off(self):
        self.index_mode = False
        self.body_render()

    def index2label(self, num):
        "given a stack number, return the index 0-9a-zA-Z, or None"

        nums = "0123456789"
        alpha_lower = "abcdefghijklmnopqrstuvwxyz"
        alpha_upper = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"

        code = nums + alpha_lower + alpha_upper

        if num < len(code):
            return code[num]

        return None

    def label2index(self, label):
        nums = "0123456789"
        alpha_lower = "abcdefghijklmnopqrstuvwxyz"
        alpha_upper = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"

        code = nums + alpha_lower + alpha_upper

        if label not in code:
            # a valid label was not entered
            return None

        index = code.index(label)

        if index >= len(self.zeitzonocities.cities):
            # a valid label was entered
            # but our list is not that long
            return None

        return index



class ZeitzonoWidgetSwitcher:
    def __init__(self, mainframe=None):
        self.widgets = {}
        self.mainframe = mainframe

    def set_mainframe(self, mainframe):
        self.mainframe = mainframe

    def set_time(self, zeitzonotime):
        self.zeitzonotime = zeitzonotime

    def set_widget(self, name, widget):
        self.widgets[name] = widget

    def set_widget_help_main(self, widget):
        self.set_widget("helpmain", widget)

    def set_widget_help_search(self, widget):
        self.set_widget("helpsearch", widget)

    def switch_widget_help_main(self):
        self._switch_widget("helpmain")

    def switch_widget_help_search(self):
        self._switch_widget("helpsearch")

    def switch_widget_main(self):
        self.widgets["main"].body_render()
        self._switch_widget("main")

    def switch_widget_search(self):
        self._switch_widget("search")

    def _switch_widget(self, name):
        self.mainframe.contents["body"] = (self.widgets[name], None)



from .ZeitzonoCities import ZeitzonoCities
from .ZeitzonoCity import ZeitzonoCity
from .ZeitzonoDB import ZeitzonoDB
from .ZeitzonoSearch import ZeitzonoSearch
from .ZeitzonoTime import ZeitzonoTime
from .ZeitzonoUrwidHelp import ZeitzonoUrwidHelp
from .ZeitzonoUrwidHelp import ZeitzonoUrwidHelpMain
from .ZeitzonoUrwidHelp import ZeitzonoUrwidHelpSearch
from .ZeitzonoUrwidMain import ZeitzonoUrwidMain
from .ZeitzonoUrwidPalette import ZeitzonoUrwidPalette
from .ZeitzonoUrwidSearch import ZeitzonoUrwidSearch
from .ZeitzonoUrwidSplashScreen import ZeitzonoUrwidSplashScreen
from .ZeitzonoWidgetSwitcher import ZeitzonoWidgetSwitcher



import urwid
import base64
import bz2
import re


class ZeitzonoUrwidSplashScreen(urwid.WidgetWrap):
    _selectable = True

    def __init__(self, zeitzonowidgetswitcher, version, dbversion):
        self.zeitzonowidgetswitcher = zeitzonowidgetswitcher
        b64 = """
        QlpoOTFBWSZTWZTfzlIAACH/gDoQIQBQjoBAAAJAIPQAAAQwAMMCREJ6kZMmmEKD
        Ro0GQGgkJE1HqZGjRk4oM2tgesad9SYwCDSCUQn2UrOVS38KX5rcZ6eGFO1OMxy0
        L2yZjG83yvza0wxsIyI5ZFk2uiOXCmy86IhXumSUcPJkgbOr5+Wwjhgmw7QacRrz
        dBuc3lEDqbD9u2TJKr0iDd07LfQgMIEvAa6UuFjDN21xYf7zVBcNktFVhkYzEZR/
        F3JFOFCQlN/OUg==
        """

        urltext = "zeitzono.org"

        bz2bin = base64.b64decode(b64)
        splashtext = bz2.decompress(bz2bin)
        splashtext = splashtext.decode("ascii")

        splashtext_lines = splashtext.splitlines()
        splash = []
        # import sys
        # sys.exit(splashtext_lines)
        splash_regex = re.compile(r"[+|\-/]")
        for line in splashtext_lines:
            if splash_regex.search(line):
                line_array = []
                for char in line:
                    ut = urwid.Text(char, wrap="clip")
                    if splash_regex.search(char):
                        attmap = urwid.AttrMap(ut, "splash_main_clock")
                    else:
                        attmap = urwid.AttrMap(ut, "splash_main")
                    line_array.append(attmap)
                line_col = urwid.Columns(line_array)
                splash.append(line_col)
            else:
                ut = urwid.Text(line, wrap="clip")
                attmap = urwid.AttrMap(ut, "splash_main")
                splash.append(attmap)

        splashpile = urwid.Pile(splash)

        presskeytext = "[Press any key to continue]\n"
        presskey = urwid.Text(presskeytext, wrap="clip", align="center")
        presskeymap = urwid.AttrMap(presskey, "splash_presskey")

        authortext = "by N.J. Thomas"
        author = urwid.Text(authortext, wrap="clip", align="right")
        authormap = urwid.AttrMap(author, "splash_author")

        url = urwid.Text(urltext, wrap="clip")
        urlmap = urwid.AttrMap(url, "splash_url")

        version_ut = urwid.Text(version, wrap="clip", align="right")
        versionmap = urwid.AttrMap(version_ut, "splash_version")

        dbversion_ut = urwid.Text(dbversion, wrap="clip", align="right")
        dbversionmap = urwid.AttrMap(dbversion_ut, "splash_version")

        topcols = urwid.Columns([urlmap, authormap])
        mainpile = urwid.Pile([topcols, splashpile, versionmap, dbversionmap])

        mainpad = urwid.Padding(mainpile, align="center", width=77)

        f = urwid.Filler(mainpad)

        frame = urwid.Frame(body=f, footer=presskeymap)
        super().__init__(frame)

    def keypress(self, size, key):
        self.zeitzonowidgetswitcher.switch_widget_main()



import datetime
import pytz
import tzlocal
from dateutil.relativedelta import relativedelta
from parsedatetime import parsedatetime
import warnings


class ZeitzonoTime:
    """
    a time object
    - basically keeps a time (datetime.datetime) and a timezone
    - initalized with local time and local timezone
    - can update time with new time (will be treated as utc)

    """

    def __init__(self):
        self.time = self._get_utc_now()
        self.set_tz_local()

    def set_tz(self, tz):
        tz = self._tz_to_pytz(tz)
        self.time = self.time.astimezone(tz)

    def set_tz_local(self):
        localtz = self._get_local_tz()
        self.time = self.time.astimezone(localtz)

    def set_time_now(self):
        tz = self.get_tz()
        self.time = self._get_utc_now().astimezone(tz)

    def set_time_nlp(self, englishtime):

        pdt_version = parsedatetime.VERSION_CONTEXT_STYLE
        pdt_cal = parsedatetime.Calendar(version=pdt_version)
        tzinfo = self.time.tzinfo
        parsed_dtresult, errno = pdt_cal.parseDT(englishtime, tzinfo=tzinfo)

        # parsedatetime docs are confusing, errno is supposed to be an int,
        # but in recent testing, we get some other object returned
        #
        # so we try and do the right thing no matter what we get
        if not isinstance(errno, int):
            errno = errno.dateTimeFlag

        if errno == 0:
            return False

        self.time = parsed_dtresult

        return True

    def get_time_utc(self):
        return self.time.astimezone(pytz.utc)

    def get_time(self):
        return self.time

    def get_time_str(self):
        return self._time_to_str(self.time)

    def get_tz(self):
        return self.time.tzinfo

    def get_time_tz_str(self, tz):
        return self._time_to_str(self.get_time_tz(tz))

    def get_time_tz(self, tz):
        return self.time.astimezone(self._tz_to_pytz(tz))

    def zero_sec(self):
        self.time = self.time.replace(second=0)

    def zero_min(self):
        self.time = self.time.replace(minute=0)

    def add_sec(self):
        self.time = self.time + datetime.timedelta(seconds=1)

    def sub_sec(self):
        self.time = self.time - datetime.timedelta(seconds=1)

    def add_min(self):
        self.time = self.time + datetime.timedelta(seconds=60)

    def sub_min(self):
        self.time = self.time - datetime.timedelta(seconds=60)

    def add_hour(self):
        self.time = self.time + datetime.timedelta(seconds=3600)

    def sub_hour(self):
        self.time = self.time - datetime.timedelta(seconds=3600)

    def add_qhour(self):
        self.time = self.time + datetime.timedelta(seconds=900)

    def sub_qhour(self):
        self.time = self.time - datetime.timedelta(seconds=900)

    def add_day(self):
        self.time = self.time + datetime.timedelta(days=1)

    def sub_day(self):
        self.time = self.time - datetime.timedelta(days=1)

    def add_qday(self):
        self.time = self.time + datetime.timedelta(hours=6)

    def sub_qday(self):
        self.time = self.time - datetime.timedelta(hours=6)

    def add_week(self):
        self.time = self.time + datetime.timedelta(days=7)

    def sub_week(self):
        self.time = self.time - datetime.timedelta(days=7)

    def add_month(self):
        self.time = self.time + relativedelta(months=1)

    def sub_month(self):
        self.time = self.time - relativedelta(months=1)

    def add_year(self):
        self.time = self.time + relativedelta(years=1)

    def sub_year(self):
        self.time = self.time - relativedelta(years=1)

    def __str__(self):
        return self._time_to_str(self.time)

    def _time_to_str(self, time):
        return time.strftime("%a") + " " + str(time)

    def _get_utc_now(self):
        now = datetime.datetime.utcnow()
        now = now.replace(microsecond=0)
        nowutc = now.replace(tzinfo=pytz.utc)
        return nowutc

    def _get_local_tz(self):

        # if tzlocal.get_localzone() cannot determine the local machine's time
        # zone, it returns UTC, but throws a warning
        #
        # we want to handle the warning gracefully, so we convert it to an
        # exception and catch it and return UTC ourselves

        warnings.simplefilter("error")

        try:
            localzone = tzlocal.get_localzone()
        except UserWarning:
            return pytz.utc

        return localzone

    def _tz_to_pytz(self, tz):
        if isinstance(tz, str):
            return pytz.timezone(tz)
        else:
            return tz



import urwid
import pkg_resources
import os


class ZeitzonoUrwidHelp(urwid.WidgetWrap):
    _selectable = True

    def __init__(self, zeitzonowidgetswitcher, helpfile, prev_is_main):
        self.zeitzonowidgetswitcher = zeitzonowidgetswitcher
        self.prev_is_main = prev_is_main

        helpfilepath = os.path.join("data", helpfile)
        helpmaintext_file = pkg_resources.resource_stream("Zeitzono", helpfilepath)
        helpmaintext = helpmaintext_file.read().decode("latin-1").strip()
        helpmaintext_file.close()

        helpmaintext = helpmaintext.split("\n")
        helpmaintext_u = [urwid.Text(line, wrap="clip") for line in helpmaintext]
        self.listwalker = urwid.SimpleListWalker(helpmaintext_u)
        self.listbox = urwid.ListBox(self.listwalker)

        frame = urwid.Frame(self.listbox, focus_part="body")
        super().__init__(frame)

    def go_previous_screen(self):
        if self.prev_is_main:
            self.zeitzonowidgetswitcher.switch_widget_main()
        else:
            self.zeitzonowidgetswitcher.switch_widget_search()

    def keypress(self, size, key):
        self.listbox.keypress(size, key)
        if isinstance(key, str):
            if key.isalpha():
                if key.lower() == "k":
                    self.listbox.keypress(size, "up")
                    return True
                if key.lower() == "j":
                    self.listbox.keypress(size, "down")
                    return True
                if key == "G":
                    self.listbox.keypress(size, "end")
                    return True
                if key == "g":
                    self.listbox.keypress(size, "home")
                    return True
                if (key.lower() == "q") or (key == "esc"):
                    self.go_previous_screen()
                    return True
                if (key == "backspace") or (key == "delete"):
                    self.listbox.keypress(size, "page up")
                    return True
            if key == " ":
                self.listbox.keypress(size, "page down")
                return True


class ZeitzonoUrwidHelpMain(ZeitzonoUrwidHelp):
    def __init__(self, zeitzonowidgetswitcher):
        helpfile = "helpmain.txt"
        prev_is_main = True
        super().__init__(zeitzonowidgetswitcher, helpfile, prev_is_main)


class ZeitzonoUrwidHelpSearch(ZeitzonoUrwidHelp):
    def __init__(self, zeitzonowidgetswitcher):
        helpfile = "helpsearch.txt"
        prev_is_main = False
        super().__init__(zeitzonowidgetswitcher, helpfile, prev_is_main)



class ZeitzonoUrwidPalette:
    palette = """
        main_zeitzono,          light red,              default
        main_version,           light cyan,             default
        main_base_time,         yellow,                 default
        main_base_time_current, light magenta,          default
        main_basezc,            yellow,                 default
        main_helpline,          light green,            default
        main_index,             light red,              default

        search_prompt,          light red,              default

        splash_url,             yellow,                 default
        splash_main,            light red,              default
        splash_main_clock,      dark green,             default
        splash_version,         light cyan,             default
        splash_presskey,        white,                  default
        splash_author,          light blue,             default

        numresults_str,         yellow,                 default
        numresults_num,         light magenta,          default
    """

    palette_lightbg = """
        main_zeitzono,          dark red,               default
        main_version,           dark cyan,              default
        main_base_time,         brown,                  default
        main_base_time_current, dark magenta,           default
        main_basezc,            brown,                  default
        main_helpline,          dark green,             default
        main_index,             dark red,               default

        search_prompt,          dark red,               default

        splash_url,             brown,                  default
        splash_main,            dark red,               default
        splash_main_clock,      brown,                  default
        splash_version,         dark cyan,              default
        splash_presskey,        black,                  default
        splash_author,          dark blue,              default

        numresults_str,         brown,                  default
        numresults_num,         dark magenta,           default
    """

    palette_nocolor = """
        main_zeitzono,          default,                default
        main_version,           default,                default
        main_base_time_current, default,                default
        main_base_time,         default,                default
        main_basezc,            default,                default
        main_helpline,          default,                default
        main_index,             default,                default

        search_prompt,          default,                default

        splash_url,             default,                default
        splash_main,            default,                default
        splash_main_clock,      default,                default
        splash_version,         default,                default
        splash_presskey,        default,                default
        splash_author,          default,                default

        numresults_str,         default,                default
        numresults_num,         default,                default
    """

    def __init__(self, no_color=False, lightbg=False):

        palette = self.palette

        if lightbg:
            palette = self.palette_lightbg

        if no_color:
            palette = self.palette_nocolor

        mypalette = palette.strip().splitlines()
        mypalette = [p.strip() for p in mypalette if p.strip()]
        mypalette = [p.split(",") for p in mypalette]
        mypalette = [[i.strip() for i in p] for p in mypalette]
        self.palette = mypalette

    def get_palette(self):
        return self.palette



import datetime
import pytz


class ZeitzonoCity:
    """
    a city object

    """

    def __init__(self, name, country, admin1, admin2, tz):
        self.name = name
        self.country = country
        self.admin1 = admin1
        self.admin2 = admin2
        self.tz = tz

    def get_tz(self):
        return self.tz

    def utc_offset(self):
        # return this city's UTC offset in seconds
        #
        # (assumes given time right now, so takes DST into consideration)
        #
        #
        timenow_tz = datetime.datetime.now(tz=pytz.timezone(self.tz))
        utcoffset_td = timenow_tz.utcoffset()
        return utcoffset_td.total_seconds()

    def _tolist(self):

        # some places like 'UTC' don't have an admin1
        if not self.admin1:
            return [self.name, self.country]

        # many places (eg. in China) don't have an admin2
        if not self.admin2:
            return [self.name, self.country, self.admin1]

        return [self.name, self.country, self.admin1, self.admin2]

    def __repr__(self):
        return str(self._tolist())

    def __str__(self):
        return str(" - ".join(self._tolist()))



import collections
import json
from .ZeitzonoCity import ZeitzonoCity
import copy


class ZeitzonoCities:
    "a city list object"

    def __init__(self, cities=None):
        if cities is None:
            self.cities = []
        else:
            self.cities = cities

        self.undo_stack = []
        self.redo_stack = []

    def save_state(self):
        city_copy = copy.copy(self.cities)
        self.undo_stack.append(city_copy)
        self.redo_stack = []

        if len(self.undo_stack) > 10:
            self.undo_stack.pop(0)

    def undo(self):
        if len(self.undo_stack) > 0:
            old_state = self.undo_stack.pop()
            self.redo_stack.append(self.cities)
            self.cities = old_state

    def redo(self):
        if len(self.redo_stack) > 0:
            new_state = self.redo_stack.pop()
            self.undo_stack.append(self.cities)
            self.cities = new_state

    def numcities(self):
        return len(self.cities)

    def isempty(self):
        return self.numcities() == 0

    def addcity(self, city):
        self.save_state()
        self.cities.insert(0, city)

    def addcities(self, hcities):
        self.save_state()
        self.cities = hcities.cities + self.cities

    def clear(self):
        self.save_state()
        self.cities = []
        self.nresults = None

    def del_first(self):
        return self.del_index(0)

    def del_last(self):
        self.save_state()
        if self.cities:
            return self.cities.pop()

    def del_index(self, index):
        self.save_state()
        if self.cities:
            return self.cities.pop(index)

    def _rotate(self, n):
        if self.cities:
            self.save_state()
            deck = collections.deque(self.cities)
            deck.rotate(n)
            self.cities = list(deck)

    def rotate_right(self):
        self._rotate(1)

    def rotate_left(self):
        self._rotate(-1)

    def roll_4(self):
        if self.numcities() >= 4:
            self.save_state()
            city1 = self.del_first()
            city2 = self.del_first()
            city3 = self.del_first()
            city4 = self.del_first()

            self.addcity(city1)
            self.addcity(city4)
            self.addcity(city3)
            self.addcity(city2)

    def roll_3(self):
        if self.numcities() >= 3:
            self.save_state()
            city1 = self.del_first()
            city2 = self.del_first()
            city3 = self.del_first()
            self.addcity(city1)
            self.addcity(city3)
            self.addcity(city2)

    def roll_2(self):
        if self.numcities() >= 2:
            self.save_state()
            city1 = self.del_first()
            city2 = self.del_first()
            self.addcity(city1)
            self.addcity(city2)

    def sort_utc_offset(self, reverse=False):
        self.save_state()
        self.cities.sort(key=lambda city: city.utc_offset(), reverse=not reverse)

    def __iter__(self):
        return iter(self.cities)

    def _hcity_to_dict(self, c):
        # used by self.toJSON() to serialize
        return c.__dict__

    def toJSON(self, filehandle):
        return json.dump(self.cities, filehandle, default=self._hcity_to_dict, indent=4)

    def fromJSON(self, filehandle):
        string_cities = json.load(filehandle)
        self.cities = []
        for sc in string_cities:
            hc = ZeitzonoCity(**sc)
            self.cities.append(hc)


class ZeitzonoCitySearch(ZeitzonoCities):

    "this is basically ZeitzonoCities with search results"

    def __init__(self, cities=None, results=0):
        super().__init__(cities=cities)

        self.results = results

    def numresults(self):
        return self.results

    def clear(self):
        super.clear()
        self.results = 0



from setuptools import setup

setup(
    name="zeitzono",
    version="v0.7.0",
    packages=["Zeitzono"],
    scripts=["bin/zeitzono"],
    install_requires=["python-dateutil", "parsedatetime", "pytz", "tzlocal", "urwid"],
    zip_safe=False,
    package_data={"Zeitzono": ["data/*"]},
    license="BSD",
    data_files=[('share/man/man1', ['man/zeitzono.1']),
                ('share/doc/zeitzono', ['README', 'LICENSE', 'LICENSE-geonames-db'])],
    python_requires=">3.6.0",
    author="N.J. Thomas",
    author_email="info@zeitzono.org",
    platforms="POSIX",
    description="city based timezone converter with a curses (TUI) interface",
    long_description=open("README").read(),
    long_description_content_type="text/plain",
    keywords="""console curses time zone timezone
                convert conversion city cities""".split(),
    url="https://zeitzono.org/",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Environment :: Console",
        "Environment :: Console :: Curses",
        "Intended Audience :: Other Audience",
        "Intended Audience :: System Administrators",
        "License :: OSI Approved :: BSD License",
        "Operating System :: POSIX",
        "Operating System :: Unix",
        "Programming Language :: Python :: 3",
        "Topic :: Office/Business :: Scheduling",
        "Topic :: Utilities",
    ],
)



import lzma
import pkg_resources
import random

DB_FILE = "data/zeitzono.db"

# if a search returns more than MAX_LIMIT_AMT results
# only return at most MAX_LIMIT_AMT
MAX_LIMIT_AMT = 100


class ZeitzonoDB:

    # search types
    SEARCH_CITY_ALL = 0   # used to search all city fields
    SEARCH_CITY_NAME = 1  # used to search city names only
    SEARCH_CADMIN = 2     # used to search against country/admin codes

    def __init__(self):

        dbfile = pkg_resources.resource_filename("Zeitzono", DB_FILE)

        # - db contains a list of dictionaries,
        #   each dictionary has the following keys:
        #
        #     name      - the name of the city (lower case)
        #     ascii     - the ascii name of the city
        #     asciil    - the ascii name of the city (lower case)
        #     alt       - alternate names for the city (lower case)
        #     cc        - the country that the city is inc
        #     ccl       - the country that the city is in (lower case)
        #     admin1    - city's admin1 code
        #     admin1l   - city's admin1 code (lower case)
        #     admin2    - city's admin2 code
        #     admin2l   - city's admin2 code (lower case)
        #     pop       - the population of the city (used for sorting)
        #     tz        - the city's timezone
        #
        # - for stuff we don't display and only search against, we store
        #   only the lower case version
        #
        # - for stuff we both display and search against, we store both
        #
        self.db = self._db_load(dbfile)

        # this is set by db_search
        # it contains the number of matches returned by a search
        self.matches = []

        # - this is set by db_search
        # - it contains the number of matches returned by a search
        # - note that count may not match len(self.matches)
        #   because we truncate at MAX_LIMIT_AMT matches
        self.numresults = 0

    def _db_load(self, dbfile):
        # the Zeitzono db is just a compressed TSV file that we open and return
        # as a list of dictionaries

        dbfh = lzma.open(dbfile, mode="rt")
        db = []

        for line in dbfh:
            db.append(self._tsv_line_2_dict(line.strip()))

        return db

    def _tsv_line_2_dict(self, tsvline):
        # given a city entry
        # return a dict that contains things we need to search or display later

        city = tsvline.strip().split("\t")
        cityd = {}

        cityd["name"] = city[0].lower()
        cityd["ascii"] = city[1]
        cityd["asciil"] = cityd["ascii"].lower()
        cityd["alt"] = city[2].lower()
        cityd["cc"] = city[3]
        cityd["ccl"] = cityd["cc"].lower()
        cityd["admin1"] = city[4]
        cityd["admin1l"] = cityd["admin1"].lower()
        cityd["admin2"] = city[5]
        cityd["admin2l"] = cityd["admin2"].lower()
        cityd["pop"] = int(city[6])
        cityd["tz"] = city[7]

        # some null admin1/admin2 listings are labeled like this,
        # so we have to null them out manually
        if cityd["admin1"] == "00":
            cityd["admin1"] = ""
            cityd["admin1l"] = ""
        if cityd["admin2"] == "00":
            cityd["admin2"] = ""
            cityd["admin2l"] = ""

        return cityd

    def _search_parser(self, search):
        # given a search string, will tokenize and return a list of tuples
        #
        #   (pattern, SEARCH_TYPE) - pattern to search for
        #                            and search type const
        #
        patterns = []
        tokens = search.strip().split()
        for tok in tokens:
            if tok.startswith(":"):
                if len(tok) > 1:
                    patterns.append((tok[1:].lower(), self.SEARCH_CADMIN))
            elif tok.startswith("'"):
                if len(tok) > 1:
                    patterns.append((tok[1:].lower(), self.SEARCH_CITY_NAME))
            else:
                patterns.append((tok.lower(), self.SEARCH_CITY_ALL))
        return patterns

    def _city_name_only_search_match(self, pattern, city):
        # return True if pattern matches city name only
        if pattern in city["name"]:
            return True
        elif pattern in city["asciil"]:
            return True
        return False

    def _city_search_match(self, pattern, city):
        # return True if pattern matches city name of altname
        if self._city_name_only_search_match(pattern, city):
            return True
        elif pattern in city["alt"]:
            return True
        return False

    def _cadmin_search_match(self, pattern, city):
        # return True if pattern matches cc or admin
        if pattern in city["ccl"]:
            return True
        elif pattern in city["admin1l"]:
            return True
        elif pattern in city["admin2l"]:
            return True
        return False

    def _sort_matches(self, matches, limit=None):
        # given a list of matches (indexes into the database)
        # return a sorted list of cities
        #
        # will always truncate to MAX_LIMIT_AMT searches
        smatches = sorted(matches, key=lambda i: self.db[i]["pop"])

        maxresults = MAX_LIMIT_AMT
        if (limit is not None) and (limit < MAX_LIMIT_AMT):
            maxresults = limit

        if (limit is not None) and (limit < MAX_LIMIT_AMT):
            maxresults = limit

        self.numresults = len(matches)

        if self.numresults > maxresults:
            smatches = smatches[-maxresults:]

        self.matches = smatches

    def db_search(self, search, limit=None):
        # will return: list, boolean
        # list contains numbers, which correspond to cities in db
        # if boolean is true, max limit was reached
        #
        # if list is empty, no cities were found

        self._reset_search()

        # tokenize search string
        patterns = self._search_parser(search)

        # need at least one city to search for
        if not patterns:
            return [], False

        matched = range(len(self.db))

        # search through with patterns
        for pattern, search_type in patterns:
            newmatched = []

            for match in matched:
                if search_type == self.SEARCH_CITY_ALL:
                    if self._city_search_match(pattern, self.db[match]):
                        newmatched.append(match)
                elif search_type == self.SEARCH_CITY_NAME:
                    if self._city_name_only_search_match(pattern, self.db[match]):
                        newmatched.append(match)
                elif search_type == self.SEARCH_CADMIN:
                    if self._cadmin_search_match(pattern, self.db[match]):
                        newmatched.append(match)

            if not newmatched:
                return [], False

            matched = newmatched

        self._sort_matches(matched, limit)

    def _reset_search(self):
        self.matches = []
        self.numresults = 0

    def random_cities(self, count):
        numrand = count if count < MAX_LIMIT_AMT else MAX_LIMIT_AMT
        population = range(len(self.db))
        randmatches = random.choices(population, k=numrand)
        self._sort_matches(randmatches)

    def match_cities(self):
        # generator that prints returns matched cities
        for m in self.matches:
            yield self.db[m]



import urwid
from .ZeitzonoSearch import ZeitzonoSearch


class ZeitzonoUrwidSearch(urwid.WidgetWrap):
    _selectable = True

    def __init__(self, zeitzonowidgetswitcher, maincities, version, db):
        self.zeitzonowidgetswitcher = zeitzonowidgetswitcher
        self.bodypile = urwid.Pile([])
        self.hsearch = ZeitzonoSearch(db)
        self.maincities = maincities
        self.results = None
        self.prompt = urwid.Edit(caption="zeitzono> ", align="left")
        prompt_attr = urwid.AttrMap(self.prompt, "search_prompt")
        self.search2pile("")

        blankline = urwid.Text("", wrap="clip")
        helpline = (
            "? - help,  0-9: add city, Enter - add bottom city, Esc - return to main"
        )
        helpline = urwid.Text(helpline, wrap="clip")
        helpline_attr = urwid.AttrMap(helpline, "main_helpline")
        footer = [blankline, helpline_attr, prompt_attr]
        footerpile = urwid.Pile(footer)

        htext = "zeitzono "
        htext_len = len(htext)
        zeitzono_ut = urwid.Text(htext, wrap="clip", align="right")
        zeitzono_ut_am = urwid.AttrMap(zeitzono_ut, "main_zeitzono")

        version_len = len(version)
        version_ut = urwid.Text(version, wrap="clip", align="right")
        version_ut_am = urwid.AttrMap(version_ut, "main_version")
        blank = urwid.Text("", align="right")
        versioncols = urwid.Columns(
            [
                ("weight", 99, blank),
                (htext_len, zeitzono_ut_am),
                (version_len, version_ut_am),
            ]
        )

        self.bodypilefiller = urwid.Filler(self.bodypile, valign="bottom")
        self.frame = urwid.Frame(
            self.bodypilefiller,
            header=versioncols,
            footer=footerpile,
            focus_part="footer",
        )
        urwid.connect_signal(self.prompt, "change", self.input_handler)
        super().__init__(self.frame)

    def input_handler(self, widget, newtext):
        if newtext:
            lastchar = newtext[-1]
            if lastchar == "?":
                return False
            if lastchar.isdigit():
                lastint = int(lastchar)
                num = self.results.numcities()
                if (num > 0) and (lastint + 1 <= num):
                    index = -1 - lastint
                    self.maincities.addcity(self.results.cities[index])
                    self.zeitzonowidgetswitcher.switch_widget_main()
        self.search2pile(newtext)

    def keypress(self, size, key):
        self.prompt.keypress((size[0],), key)
        if isinstance(key, str):
            if key == "ctrl g":
                self.prompt.set_edit_text("")
            if key == "?":
                self.prompt.set_edit_text("")
                self.zeitzonowidgetswitcher.switch_widget_help_search()
            if key == "enter":
                if self.results is not None:
                    if self.results.numcities() >= 1:
                        self.maincities.addcity(self.results.cities[-1])
                self.zeitzonowidgetswitcher.switch_widget_main()
            if key == "esc":
                self.results.clear()
                self.zeitzonowidgetswitcher.switch_widget_main()

    def _item2pileitem(self, item):
        options = ("weight", 1)
        pitem = (item, options)
        return pitem

    def _get_max_capacity(self):
        # this is kludgy because Urwid doesn't give us a really great way to
        # get the size of some widget
        cols, rows = urwid.raw_display.Screen().get_cols_rows()
        maxrows = rows - 6
        return maxrows

    def search2pile(self, terms):
        newlist = []

        terms = terms.strip()
        hcities = self.hsearch.search(terms, limit=self._get_max_capacity())
        numresults = hcities.numresults()

        numresults_t = urwid.Text("NUMRESULTS: ")
        numresultst_map = urwid.AttrMap(numresults_t, "numresults_str")
        numresults_num_t = urwid.Text("%s" % str(numresults))
        numresults_num_map = urwid.AttrMap(numresults_num_t, "numresults_num")
        numresultscols = urwid.Columns(
            [("pack", numresultst_map), ("pack", numresults_num_map)]
        )

        for idx, city in enumerate(hcities):
            stacknum = hcities.numcities() - idx - 1
            citys = str(city)
            if stacknum > 9:
                cityt = urwid.Text("   " + citys, wrap="clip", align="left")
            else:
                cityt = urwid.Text(
                    str(stacknum) + ": " + citys, wrap="clip", align="left"
                )
            newlist.append(cityt)
        newlist.append(urwid.Text("", align="right"))
        newlist.append(numresultscols)

        newlist = [self._item2pileitem(item) for item in newlist]

        self.bodypile.contents[:] = newlist
        self.results = hcities



from .ZeitzonoCities import ZeitzonoCitySearch
from .ZeitzonoCity import ZeitzonoCity


class ZeitzonoSearch:
    """
    will search for city names
    returns Cities object with number of results found

    """

    def __init__(self, db):
        self.zdb = db

    def _format_db_results(self):

        cities = []
        for city in self.zdb.match_cities():
            name = city["ascii"]
            country = city["cc"]
            admin1 = city["admin1"]
            admin2 = city["admin2"]
            tz = city["tz"]
            city = ZeitzonoCity(name, country, admin1, admin2, tz)
            cities.append(city)

        return ZeitzonoCitySearch(cities=cities, results=self.zdb.numresults)

    def search(self, search, limit=None):
        if not search:
            return ZeitzonoCitySearch()

        self.zdb.db_search(search, limit)

        numresults = self.zdb.numresults

        if numresults <= 0:
            return ZeitzonoCitySearch()

        return self._format_db_results()

    def random1(self):
        self.zdb.random_cities(1)
        return self._format_db_results()

    def random10(self):
        self.zdb.random_cities(10)
        return self._format_db_results()
